@@ -1,5 +1,5 @@
 
-def _collapse(pattern: str) -> str:
+def collapse(pattern: str) -> str:
     out = ""
     i = 0
     while i < len(pattern):
@@ -51,4 +51,4 @@ class Glob(object):
     def match(self, s: str) -> bool:
         return _match(self._pattern, s)
 def compile(pattern: str) -> Glob:
-    return Glob(_collapse(pattern))
+    return Glob(collapse(pattern))
